import aiofiles
import aiohttp
import asyncio
import async_timeout
import os
import numpy as np
import pandas as pd

async def download_coroutine(session, url, group, sample, source, evalue=5, identity=60, length=15):
    with async_timeout.timeout(None):
        async with session.get(url, params = {'source': source,'evalue': evalue, 'identity': identity, 'length': length, 'type': 'all'}) as response:
            filename = '%(group)s_%(sample)s.tsv' % \
                {"group": group, "sample": sample}

            async with aiofiles.open(filename, 'wb') as fd:
                while True:
                    chunk = await response.content.read(10240)
                    if not chunk:
                        break
                    await fd.write(chunk)
            return await response.release()


async def download_geneonly_coroutine(session, url, study, idList, source, evalue=5, identity=60, length=15):
    with async_timeout.timeout(None):
        params_ids = {'id': mgID for mgID in idList}
        params_other =  {'source': source,'evalue': evalue, 'identity': identity, 'length': length}
        params = params_ids.copy()
        params.update(params_other)
        async with session.get(url, params = params) as response:
            filename = '%(study)s_function_counts.biom' % \
                {"study": study}

            async with aiofiles.open(filename, 'wb') as fd:
                while True:
                    chunk = await response.content.read(10240)
                    if not chunk:
                        break
                    await fd.write(chunk)
            return await response.release()
'''
async def main(loop, study, metagenomeGroupDict, source, evalue, identity, length):
    metagenomeGroupDF = pd.DataFrame({'metagenome ID' : list(metagenomeGroupDict.keys()),'group' : list(metagenomeGroupDict.values())})
    metagenomeGroupDF['sample'] = metagenomeGroupDF.groupby('group').cumcount()+1

    urls = ["http://api.metagenomics.anl.gov/annotation/similarity/" + str(id) for id in metagenomeGroupDF['metagenome ID']]
    geneOnlySeries = pd.Series(['all', 'gene-only', 'counts-all'], index=['metagenome ID','group','sample'])
    #metagenomeGroupDF.append(geneOnlySeries, ignore_index=True)
    geneOnlyURL = 'http://api.metagenomics.anl.gov/matrix/function'# + ''.join()id=mgm4447943.3&id=mgm4447192.3&id=mgm4447102.3  

    async with aiohttp.ClientSession(loop=loop) as session:
'''
       # for i in range(len(metagenomeGroupDF)):
           #await download_coroutine(session, urls[i], metagenomeGroupDF['group'][i], metagenomeGroupDF['sample'][i], source, evalue, identity, length)
        #await download_geneonly_coroutine(session, geneOnlyURL, study, metagenomeGroupDF['metagenome ID'], source, evalue, identity, length)
'''
    tasks = [asyncio.ensure_future(download_coroutine(session, urls[i], metagenomeGroupDF['group'][i], metagenomeGroupDF['sample'][i], source, evalue, identity, length)) for i in range(len(metagenomeGroupDF))]
'''

def ensure_all_futures(loop, study, metagenomeGroupDict, source, evalue, identity, length):
   metagenomeGroupDF = pd.DataFrame({'metagenome ID' : list(metagenomeGroupDict.keys()),'group' : list(metagenomeGroupDict.values())})
   metagenomeGroupDF['sample'] = metagenomeGroupDF.groupby('group').cumcount()+1
   urls = ["http://api.metagenomics.anl.gov/annotation/similarity/" + str(id) for id in metagenomeGroupDF['metagenome ID']]
   geneOnlyURL = 'http://api.metagenomics.anl.gov/matrix/function'# + ''.join()id=mgm4447943.3&id=mgm4447192.3&id=mgm4447102.3  

   with aiohttp.ClientSession(loop=loop) as session:
      tasks = [asyncio.ensure_future(download_coroutine(session, urls[i], metagenomeGroupDF['group'][i], metagenomeGroupDF['sample'][i], source, evalue, identity, length)) for i in range(len(metagenomeGroupDF))]
      tasks.append(asyncio.ensure_future(download_geneonly_coroutine(session, geneOnlyURL, study, metagenomeGroupDF['metagenome ID'], source, evalue, identity, length)))
      return tasks

def download_study_data(study, metagenomeGroupDict, source, evalue, identity, length):
#if __name__ == '__main__':
    metagenomeGroupDF = pd.DataFrame({'metagenome ID' : list(metagenomeGroupDict.keys()),'group' : list(metagenomeGroupDict.values())})
    metagenomeGroupDF['sample'] = metagenomeGroupDF.groupby('group').cumcount()+1
    loop = asyncio.get_event_loop()
#   async with aiohttp.ClientSession(loop=loop) as session:
#       tasks = [asyncio.ensure_future(download_coroutine(session, urls[i], metagenomeGroupDF['group'][i], metagenomeGroupDF['sample'][i], source, evalue, identity, length)) for i in range(len(metagenomeGroupDF))]
#       tasks.append(asyncio.ensure_future(download_geneonly_coroutine(session, geneOnlyURL, study, metagenomeGroupDF['metagenome ID'], source, evalue, identity, length)))
    #loop.run_until_complete(main(loop, study, metagenomeGroupDict, source, evalue, identity, length))
    loop.run_until_complete(asyncio.wait(ensure_all_futures(loop, study, metagenomeGroupDict, source, evalue, identity, length)))





import aiofiles
import aiohttp
import asyncio
import async_timeout
import os
import numpy as np
import pandas as pd
import os
pd.options.mode.chained_assignment = None

def create_path(path):
    """
    Summary: creates a path, with error checking

    Args:
        path (str): name of folder/directory to be created

    Returns:
        None, creates folder
    """

    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise

async def download_coroutine(session, url, group, sample, source, evalue, identity, length):
    """
    Summary:

    Args:

    Returns:
    """

    async with session.get(url, params = {'source': source,'evalue': evalue, 'identity': identity, 'length': length, 'type': 'all'}, timeout=None) as response:
        filename = '%(group)s_%(sample)s.tsv' % \
            {"group": group, "sample": sample}
        print('reading '+ str(filename))

        async with aiofiles.open(filename, 'wb') as fd:
            print('writing ' + str(filename))
            while True:
                chunk = await response.content.read(1024)
                if not chunk:
                    break
                await fd.write(chunk)
        return await response.release()


async def download_geneonly_coroutine(session, url, study, idList, source, evalue, identity, length):
    """
    Summary:

    Args:

    Returns:
    """
    params_ids = {'id': mgID for mgID in idList}
    params_other =  {'source': source,'evalue': evalue, 'identity': identity, 'length': length}
    params = params_ids.copy()
    params.update(params_other)
    geneOnlyURL = 'http://api.metagenomics.anl.gov/matrix/function?id=' + '&id='.join(idList)+ '&source=' + str(source)+ '&evalue='  + str(evalue) + '&identity=' + str(identity) + '&length=' + str(length)

    async with session.get(url, params = params, timeout=None) as response:
        filename = '%(study)s_function_counts.biom' % \
            {"study": study}

        async with aiofiles.open(filename, 'wb') as fd:
            while True:
                chunk = await response.content.read(128)
                if not chunk:
                    break
                await fd.write(chunk)
        return await response.release()





async def download_all(urls, geneOnlyUrl, study, metagenomeGroupDF, source, evalue, identity, length):
    """Launch requests for all web pages."""
    tasks = []
    idList = metagenomeGroupDF['metagenome ID']
    groupList = metagenomeGroupDF['group']
    sampleList = metagenomeGroupDF['sample']
    create_path(str(study))
    os.chdir(str(study))
    async with aiohttp.ClientSession() as session:
        for i in range(len(urls)):
            task = asyncio.ensure_future(download_coroutine(session, urls[i], groupList[i], sampleList[i], source, 5, 60, 15))
            tasks.append(task) # create list of tasks
        tasks.append(asyncio.ensure_future(download_geneonly_coroutine(session, geneOnlyUrl, study, idList, source, 5, 60, 15)))
        _ = await asyncio.gather(*tasks) # gather task responses


def get_all_async(study, metagenomeGroupDict, source, evalue=5, identity=60, length=15):
    """
    Summary: gets all metagenmic annotations for a study from the MG-RAST servers and stores as CSV

    Args:
        study (str)
        metagenomeGroupDict (dict)
        source (str)
        evalue (int)
        identity (int)
        length (int)

    Returns:
        None, saves all annotations in folder named study
    """
    metagenomeGroupDF = pd.DataFrame({'metagenome ID' : list(metagenomeGroupDict.keys()),'group' : list(metagenomeGroupDict.values())})
    metagenomeGroupDF['sample'] = metagenomeGroupDF.groupby('group').cumcount()+1
    urls = ["http://api.metagenomics.anl.gov/annotation/similarity/" + str(id) for id in metagenomeGroupDF['metagenome ID']]
    geneOnlyUrl = 'http://api.metagenomics.anl.gov/matrix/function'
    loop = asyncio.get_event_loop() # event loop
    future = asyncio.ensure_future(download_all(urls, geneOnlyUrl, study, metagenomeGroupDF, source, evalue, identity, length)) # tasks to do
    loop.run_until_complete(future) # loop until done
    os.chdir('..')


if __name__ == '__main__':
    get_all_async('lagoon study', {'mgm4739968.3':'nador_lagoon', 'mgm4739969.3':'oualidia_lagoon', 'mgm4739970.3': 'oualidia_lagoon', 'mgm4739971.3':'nador_lagoon'}, 'RefSeq', evalue=5, identity=60, length=15)



import glob
import math
import numpy as np
import pandas as pd
import sys
import os
import errno
import itertools
from skbio.stats import subsample_counts
from functools import reduce
from Bio import Entrez
import time
import seaborn as sns
import re
#import requests
from scipy import stats
from bokeh.charts import Donut, show, output_file
from bokeh.layouts import row, column
from bokeh.palettes import Set3
############################################################################################
'''
DETERMING THE RELATIONSHIP BETWEEN CHANGE IN DIVERSITY AND CHANGE IN FUNCTION











'''
############################################################################################

pd.options.mode.chained_assignment = None
###SEARCH BY GENE###
#takes dataframe
def search_by_gene(query, dfCounts):
    """
    Summary: gets the subset of a dataframe by the query of interest
    Args:
        query (str): search term of interest, use [space] for OR & | [pipe] for AND
        dfCounts (pandas.DataFrame): counts dataframe
    Returns:
        geneSorted (pandas.DataFrame): dataframe for genes containing query
    """

    queryList = query.replace(" ", "|").replace("+", " ")
    geneSorted = dfCounts[dfCounts['gene function'].str.contains(queryList)]
    return geneSorted

def merge_dfs2(ldf, rdf):
    """
    Summary: merges two counts (with hierarchy)
    Args:
        ldf (pandas.DataFrame): left dataframe
        rdf (pandas.DataFrame): right dataframe
    Returns:
        pd.merge(ldf, rdf) (pandas.DataFrame): merged dataframe
    """

    print('yep, again...')
    return ldf.merge(rdf, how='outer', on=['gene function', 'organism', 'phylum', 'class', 'order', 'family'])

def merge_all_counts(studyName):
    """
    Summary: merges all counts w/ hierarchy
    Args:
        studyName (str): study name/directory name (ideally one and the same)
    Returns:
        mergedCounts (pandas.DataFrame): merged dataframe
    """

    annotatedCountsList = glob.glob(studyName + '/norm_taxonomy/*norm_taxonomy.csv')
    annotatedDFsList = [pd.read_csv(count) for count in annotatedCountsList]
    [annDF.pop('Unnamed: 0') for annDF in annotatedDFsList]
    mergedCounts  = reduce(merge_dfs2, annotatedDFsList)
    mergedCounts.fillna(0)
    mergedCounts.to_csv(str(studyName) + '_allcounts.csv')
    return mergedCounts

def split_hierarchically(geneSorted):
    """
    Summary: breaks out the Entrez lineage column into separate columns for each taxonomic level
    Args:
        geneSorted (pandas.DataFrame): annotated counts with column for hierarchy
    Returns:
        taxList (list): list of counts dataframes, each at a different taxonomic level
    """

    taxList = []
    taxList.append(geneSorted.groupby('phylum').sum())
    taxList.append(geneSorted.groupby('class').sum())
    taxList.append(geneSorted.groupby('order').sum())
    taxList.append(geneSorted.groupby('family').sum())
    return taxList

def subset_by_gene(studyName, geneName, testMerge):
    """
    Summary: subsets counts dataframe by query (gene name/gene family name)
    Args:
        studyName (str): name of study directory
        geneName (str): query of interest
        testMerge (pandas.DataFrame): merged annotated counts
    Returns:
        geneDF (pandas.DataFrame): dataframe of counts only for gene/query of interest
    """

    os.chdir(studyName)
    geneDF = search_by_gene(geneName, testMerge)
    geneDF.to_csv(str(geneName)+'_'+'allcounts.csv')
    os.chdir('..')
    return geneDF

##############################################################################################
####PLOT RENYI PROFILE###
#input is a data frame of values with rows representing taxa (in the individual columns) and columns representing the individual samples, as labelled with [group]_[samplenumber]

###MAKE COUNTS PROPORTIONAL (FOR RENYI PROFILE ETC)###
def make_counts_proportional(counts):
    """
    Summary: converts numerical counts to fractional representation
    Args:
        counts (pandas.Series): counts
    Returns:
        counts (pandas.Series): counts represented proportionally
    """

    counts = counts/counts.sum()
    return counts
###CALCULATE RENYI ENTROPY###
def renyientropy(px, alpha):
    """
    Summary: calculates renyi entropy of array px at alpha
    Args:
        px (array-like object): distribution  of counts
        alpha (): alpha value at which renyi is to be calculated 
    Returns:
        renyi (): renyi entropy of px at alpha 
    """

    if alpha < 0:
        raise ValueError("alpha must be a non-negative real number")
    elif alpha == 0:
        renyi = np.log2(len(px[px > 0])+.00000000000001)
    elif alpha == 1:
        renyi = -np.sum(np.nan_to_num(px*np.log2(px+.00000000000001)))
    elif alpha == np.inf:
        renyi = -np.log2(px.max())
    else:
        renyi = 1/(1-alpha) * np.log2((px**alpha).sum())
    return renyi
#####CALCULATE RENYI PROFILE#####
def calculate_renyi_profile(lvlCounts):
    """
    Summary: calculates renyi profile of a counts distribution
    Args:
        lvlCounts (array-like object): counts distribution
    Returns:
        profile_values (list): renyi profile of lvlCounts
    """

    profile_values = []
    if lvlCounts.sum() == 0:
        profile_values = [0,0,0,0,0,0,0,0,0,0,0]
    else:
        lvlCounts = lvlCounts/lvlCounts.sum()
        counts_distribution = lvlCounts

        profile_values.append(renyientropy(counts_distribution, 0))
        profile_values.append(renyientropy(counts_distribution, .25))
        profile_values.append(renyientropy(counts_distribution, .5))
        profile_values.append(renyientropy(counts_distribution, 1))
        profile_values.append(renyientropy(counts_distribution, 2))
        profile_values.append(renyientropy(counts_distribution, 4))
        profile_values.append(renyientropy(counts_distribution, 8))
        profile_values.append(renyientropy(counts_distribution, 16))
        profile_values.append(renyientropy(counts_distribution, 32))
        profile_values.append(renyientropy(counts_distribution, 64))
        profile_values.append(renyientropy(counts_distribution, np.inf))
    return profile_values


###RENYI PROFILE VIA SEABORN###(10 AUG 2017)###
def renyi_profile_of_level(levelName,countsDF):
    """
    Summary: calculate renyi profile at a taxonomic level
    Args:
        levelName (str): phlyum, class, order, or family
        countsDF (pandas.DataFrame): dataframe of counts at taxonomic level
    Returns:
    """

    alphasList = []
    renyiProfileList = []
    bigGroupList = []
    #write similarly to make_renyiprofile_list
    for i in range(len(countsDF.columns.tolist())):
        renyiProfileList.insert(i, calculate_renyi_profile(countsDF.ix[:, i]))
        alphasList.insert(i, ['0','.25','.5','1','2','4','8','16','32','64', 'Inf'])
        bigGroupList.insert(i , [countsDF.columns[i].split('_')[0]] * 11)

    flatAlphaList = list(itertools.chain.from_iterable(alphasList))
    flatGroupList = list(itertools.chain.from_iterable(bigGroupList))
    flatRenyiList = list(itertools.chain.from_iterable(renyiProfileList))
    levelList = [levelName] * len(flatAlphaList)
    countsDF.columns.str.split('_').str[0]

    renyiDF = pd.DataFrame({'level': levelList,'group':flatGroupList,'alpha':flatAlphaList,'H_alpha':flatRenyiList})
    renyiDF.to_csv(levelName +'_renyi_profile.csv')

    return renyiDF

def plot_renyi_PCOF(geneList):
    """
    Summary: plots all four renyi profiles for a gene of interest: phylum, class, order, and family
    Args:
        geneList (list): list of dataframes, one in each taxonomic level
    Returns:
    """

    levelsList = ['phylum','class','order','family']
    renyiDFList = [renyi_profile_of_level(levelsList[i], geneList[i]) for i in range(len(geneList))]
    mergedRenyiDF = pd.concat(renyiDFList)
    sns.set(style='whitegrid')
    sns.factorplot(x="alpha", y="H_alpha", col="level", hue="group", data= mergedRenyiDF,
                   col_wrap=4, ci=None, palette="muted", size=15, aspect=.85)
    return




'''
##step 1: create a list of four separate dataframes for counts summed by phylum, class, order, and family:

##SUBSET DATAFRAMES
def get_groups_list(dataCounts):
    """
    Summary:
    Args:
    Returns:
    """

    grpsList =  dataCounts.columns.str.split("_").str[0].unique()
    return grpsList
'''
def average_by_groups(grpsList, taxListLevel):
    """
    Summary:
    Args:
    Returns:
    """

    grpsMeanList = []
    for i in range(len(grpsList.tolist())):
        grpsMeanList.insert(i, taxListLevel.filter(like=grpsList.tolist()[i]).mean(1))
    return grpsMeanList

def sem_by_groups(grpsList, taxListLevel):
    """
    Summary:
    Args:
    Returns:
    """

    grpsSemList = []
    for i in range(len(grpsList.tolist())):
        grpsSemList.insert(i, taxListLevel.filter(like=grpsList.tolist()[i]).sem(1))
    return grpsSemList

def avg_renyi_by_groups(grpsList, taxListLevel):
    """
    Summary:
    Args:
    Returns:
    """

    grpsMeanList = []
    for i in range(len(grpsList.tolist())):
        grpsMeanList.insert(i, np.average(make_renyiprofile_list(taxListLevel.filter(like=grpsList.tolist()[i])), axis=0))
    return grpsMeanList

def sem_renyi_by_groups(grpsList, taxListLevel):
    """
    Summary:
    Args:
    Returns:
    """

    grpsSemList = []
    for i in range(len(grpsList.tolist())):
        grpsSemList.insert(i, stats.sem(make_renyiprofile_list(taxListLevel.filter(like=grpsList.tolist()[i]))))
    return grpsSemList
'''
def plot_hierarchical_renyi(taxList, geneName):
    """
    Summary:
    Args:
    Returns:
    """

    grpsList = get_groups_list(taxList[0])

    #!/usr/bin/env python
    from scipy import stats

    import numpy as np
    import matplotlib.pyplot as plt
    avgRenyiList = []
    semRenyiList = []
    avgRenyiList1 = []
    semRenyiList1 = []
    avgRenyiList2 = []
    semRenyiList2 = []
    avgRenyiList3 = []
    semRenyiList3 = []
    # example data
    for i in range(len(grpsList.tolist())):
        avgRenyiList.insert(i, avg_renyi_by_groups(grpsList,taxList[0])[i])
        semRenyiList.insert(i, sem_renyi_by_groups(grpsList,taxList[0])[i])

        avgRenyiList1.insert(i, avg_renyi_by_groups(grpsList,taxList[1])[i])
        semRenyiList1.insert(i, sem_renyi_by_groups(grpsList,taxList[1])[i])

        avgRenyiList2.insert(i, avg_renyi_by_groups(grpsList,taxList[2])[i])
        semRenyiList2.insert(i, sem_renyi_by_groups(grpsList,taxList[2])[i])

        avgRenyiList3.insert(i, avg_renyi_by_groups(grpsList,taxList[3])[i])
        semRenyiList3.insert(i, sem_renyi_by_groups(grpsList,taxList[3])[i])






    renyi_vals = np.asarray([0, .25, .5, 1, 2, 4, 8, 16, 32, 64, np.inf])

    # First illustrate basic pyplot interface, using defaults where possible.

    import matplotlib.gridspec as gridspec
    fig = plt.figure(dpi=1600)
    # Now switch to a more OO interface to exercise more features.
    gs1 = gridspec.GridSpec(4, 1)


    ax = fig.add_subplot(gs1[0])
    for i in range(len(grpsList.tolist())):
        ax.errorbar(renyi_vals, avgRenyiList[i], label=grpsList.tolist()[i], yerr=semRenyiList[i])
    ax.set_xscale('log')
    ax.set_title('phylum')

    # With 4 subplots, reduce the number of axis ticks to avoid crowding.

    ax = fig.add_subplot(gs1[1])
    for i in range(len(grpsList.tolist())):
        ax.errorbar(renyi_vals, avgRenyiList1[i], label=grpsList.tolist()[i], yerr=semRenyiList1[i])
    ax.set_xscale('log')
    ax.set_title('class')

    ax = fig.add_subplot(gs1[2])
    for i in range(len(grpsList.tolist())):
        ax.errorbar(renyi_vals, avgRenyiList2[i], label=grpsList.tolist()[i], yerr=semRenyiList2[i])
    ax.set_xscale('log')
    ax.set_title('order')

    ax = fig.add_subplot(gs1[3])
    for i in range(len(grpsList.tolist())):
        ax.errorbar(renyi_vals, avgRenyiList2[i], label=grpsList.tolist()[i], yerr=semRenyiList2[i])
    # Here we have to be careful to keep all y values positive:
    ax.set_xscale('log')
    ax.set_title('family')

    fig.suptitle('Diversity Measures of Organisms with ' + geneName)
    fig.set_size_inches(35, 25, forward=True)
    plt.legend(loc='upper center')
    plt.show()
    return
'''
'''
#step 2.1: create list of distributions from each column of dataframe
def make_renyiprofile_list(countsDF):
    """
    Summary:
    Args:
    Returns:
    """

    distributionsList = []
    renyiProfileList = []
    for i in range(len(countsDF.columns.tolist())):

        renyiProfileList.insert(i, np.array(calculate_renyi_profile(countsDF.ix[:, i])))
    return renyiProfileList


#Step 3: Calculate Renyi profile lists for all four dataframes in the taxa list
def calculate_hierarchical_renyi(taxList):
    """
    Summary:
    Args:
    Returns:
    """

    hierRenyiList = []
    for i in range(len(taxList)):
      hierRenyiList.insert(i,make_renyiprofile_list(taxList[i]))
    return hierRenyiList
'''

###CALCULATE DOUGHNUT (DONUT) CHARTS ####


def fractionalise_by_column(df):
    """
    Summary: converts counts to fractions
    Args:
        df (pandas.DataFrame): counts dataframe
    Returns:
        df (pandas.DataFrame): counts as percentages
    """

    for column in df:
        if(df[column].dtype == np.float64 or df[column].dtype == np.int64):
            df[column] = df[column]/(df[column].sum())
        else:
            df[column] = df[column]
    return df


def percentify_by_column(df):
    """
    Summary: converts counts to percentages
    Args:
        df (pandas.DataFrame): counts dataframe
    Returns:
        df (pandas.DataFrame): counts as percentages
    """

    for column in df:
        if(df[column].dtype == np.float64 or df[column].dtype == np.int64):
            df[column] = df[column] * 100
        else:
            df[column] = df[column]
    return df


def make_diversity_donut(geneSorted, geneName, inner, outer):
    """
    Summary: makes Donut plot at two taxonomic levels from counts dataframe
    Args:
        geneSorted (): counts data subset by gene of interest
        geneName (): gene name of interest
        inner (str): inner layer of donut plot
            outer (str): outer layer of donut plot
    Returns:
        donutList (list): outputs donut plots for each group at levels outer and inner 
    """

    cols = pd.Series([c for c in geneSorted.columns if '_'  in  c.lower()])
    grpsList = cols.str.split("_").str[0].unique()

    geneAvg = average_by_groups(grpsList, geneSorted)
    for i in range(len(geneAvg)):
        geneSorted[str(grpsList[i])] = geneAvg[i]
    cols = pd.Series([c for c in geneSorted.columns if '_' not in  c.lower()])
    geneSortedMean = geneSorted[cols]
    geneMeanList = []
    grpsList = grpsList.tolist()
    for group in grpsList:
        geneMeanList.insert(grpsList.index(group),  geneSortedMean[geneSortedMean[group] > 0])
    donutList = []
    vcColourList = ["#007892",
            "#ff6521",
            "#8c39e9",
            "#00cd1a",
            "#b900cb",
            "#b7d200",
            "#001c8e",
            "#ffd638",
            "#0045a0",
            "#efa200",
            "#41025b",
            "#a9f17a",
            "#ff92ff",
            "#548900",
            "#de0061",
            "#74f6bc",
            "#e40038",
            "#01ac95",
            "#ff65ab",
            "#008c5b",
            "#860025",
            "#00b7d5",
            "#ae4c00",
            "#82b6ff",
            "#c77300",
            "#0181c1",
            "#ffba5e",
            "#311852",
            "#d9e49f",
            "#10244e",
            "#ff8a65",
            "#003f5a",
            "#ff8492",
            "#005a25",
            "#b5aeff",
            "#5d5f00",
            "#e5d9f5",
            "#781f00",
            "#c8e2ec",
            "#500907",
            "#eaddbe",
            "#3c183b",
            "#896500",
            "#016c70",
            "#6e002f",
            "#002d1c",
            "#ffabbc",
            "#462f00",
            "#45132c",
            "#372016"]

    for k in range(len(geneMeanList)):
        donutList.insert(k, Donut(geneMeanList[k], label=[inner, outer], values=grpsList[k], palette = Set3[12], text_font_size = '1em', hover_text = grpsList[k], plot_height = 2000, plot_width = 2000, title = geneName + " by " +  inner + " and " + outer + ": group " + grpsList[k]))
    output_file(str(geneName) + "_" + str(inner) + "_" +  str(outer) + ".html")
    show(column(donutList))
    return donutList



def get_hierarchical_DD(geneSorted, geneName):
    """
    Summary: create donuts at all taxonomic levels (phylum/class, class/order, order/family) for gene of interest
    Args:
        geneSorted (pandas.DataFrame): counts data subset by gene of interest
        geneName (str): gene name
    Returns:
        None, outputs html files containing donut plots for each taxonomic level and group
    """

    print("generating for phylum -> class...")
    pcDonutList = make_diversity_donut(geneSorted, geneName, 'phylum', 'class')
    print("generating for class -> order...")
    coDonutList = make_diversity_donut(geneSorted, geneName, 'class', 'order')
    print("generating for order -> family...")
    ofDountList = make_diversity_donut(geneSorted, geneName, 'order', 'family')
    return

def get_diversity_measures(geneName, searchString, studyName):
    """
    Summary: creates renyi plot, donut, and outputs code for treemap all as visualisations of diversity for gene of interest
    Args:
        geneName (str): actual gene name
        searchString (str): query for gene of interest
        studyName (str): name of study directory
    Returns:
        Renyi plots
        Donut plots
        code to generate treemaps via Brunel
    """

    print('merging pre-processed counts:')
    mergedCounts = merge_all_counts(studyName)
    print('subsetting merged counts table by gene '+ geneName + '...')
    geneDF = subset_by_gene(studyName,searchString, mergedCounts)
    avgGeneDF = output_treemap(geneDF, geneName, studyName)
    geneList = split_hierarchically(geneDF)
    [geneList[i].to_csv('level'+ str(i+1) +'_counts.csv') for i in range(len(geneList))]

    print('''
    quantitative measure of diversity: Renyi profile by taxonomic level

                  |
                  |   ^                                    ^
                  |   |                                    |
    Diversity     | richness                            evenness
    [H_alpha(X)]  |   |                                    |
                  |   v                                    v
                  |______________________________________________
                   0                   alpha                 inf

    The Renyi Profile is a generalised version of the Shannon Entropy, where
    the Renyi entropy of order alpha:

                    log(sum(i=1->n, (p_i)^alpha))
    H_alpha(X) =    -----------------------------
                             1 - alpha

    unless alpha = 1, in which case H_alpha(X) is the Shannon Entropy:

    H_1(X) = - sum(i=1->n, p_i*log(p_i))

    If there are multiple communities, they can be ranked using the renyi profile
    (see above- vertical axis allows for partial or complete diversity ordering)

    Is the change in gene counts due to a change in diversity? This provides a quantitative answer.

    ''')
    print(geneName)
    plot_renyi_PCOF(geneList)
    print('''
    qualitative measure of diversity: Donut plot by taxonomic level

    A quantitative measure of diversity is useful for ranking diversity very easily, but specific
    changes in species can be obsucred by quantitative measures alone.

    For each taxonomic level, a Donut (hierarchical pie chart) is generated for each group.
    The widget for each table can be adjusted according to the user's needs, and then saved.
    ''')
    #get_hierarchical_DD(geneDF, geneName)
    return avgGeneDF



'''PLOTTING RENYI AND TREEMAP'''


#Renyi profile: using seaborn, reading CSV
#Donut: using Bokeh, reading CSV
#Treemap: using Brunel, reading CSV

#Renyi CSV
#input:
#output: dataframe or csv with columns for each group's renyi profile, and rows for each value in the renyi profilealpha
'''
Renyi will be plotted via seaborn
Group  alpha  H_alpha(X)
------|------|----------
      |  0   |
      |  ... |
      |  inf |

'''
#Brunel Treemap CSV
'''

%brunel data('.csv') treemap x(phylum) y(class) color(phylum) size(counts) sort(counts) label(class) tooltip(#all)
%brunel data('.csv') treemap x(class) y(order) color(class) size(counts) sort(counts) label(order) tooltip(#all)
%brunel data('.csv') treemap x(order) y(family) color(order) size(counts) sort(counts) label(family) tooltip(#all)
'''
'''







'''
#brunel command to plot: data('') x(alpha)

def output_treemap(geneSorted, geneName, studyName):
    """
    Summary: generates code to make treemaps
    Args:
        geneSorted (pandas.DataFrame): gene counts DF with hierarchy
        geneName (str): actual gene name
        studyName (str): name of study directory
    Returns:
        None, prints out code that can be copied into Jupyter as-is to generate treemaps via Brunel
    """

    cols = pd.Series([c for c in geneSorted.columns if '_'  in  c.lower()])
    grpsList = cols.str.split("_").str[0].unique()

    geneAvg = average_by_groups(grpsList, geneSorted)
    for i in range(len(geneAvg)):
        geneSorted[str(grpsList[i])] = geneAvg[i]
    cols = pd.Series([c for c in geneSorted.columns if '_' not in  c.lower()])
    geneSortedMean = geneSorted[cols]
    geneMeanList = []
    grpsList = grpsList.tolist()
    for k in range(len(geneAvg)):
        geneSortedMean[grpsList[k]] = geneAvg[k]
    donutList = []
    fileName = (studyName + '_' + geneName.replace(' ', '') + '_avg.csv')
    geneSortedMean.to_csv(fileName)
    print('''
    INSTRUCTIONS FOR GENERATING DIVERSITY TREEMAPS
    First copy the import statements and the read_csv into a cell.
    For each figure you want to generate, copy each line starting with "%brunel"  into its own  iPython/Jupyter cell:

    ''')
    variableName = geneName.replace(' ', '')
    print('''
    #copy this into a cell
    import brunel
    import pandas as pd
    {0} = pd.read_csv('{1}')
    '''.format(variableName , fileName))
    for e in range(len(grpsList)):
        print('''
        %brunel data('{0}') treemap x(phylum) y(class) color(phylum) size({1}) sort({1}) label(organism, {1}) tooltip(#all) :: width=2000, height=2000

        %brunel data('{0}') treemap x(class) y(order) color(class) size({1}) sort({1}) label(organism, {1}) tooltip(#all) :: width=2000, height=2000

        %brunel data('{0}') treemap x(order) y(family) color(order) size({1}) sort({1}) label(organism, {1}) tooltip(#all) :: width=2000, height=2000
        '''.format(variableName, grpsList[e], fileName))
        #geneMeanList[k] grpsList[k]
    return geneSortedMean

def search_by_taxa(query, level, dfCounts):
    """
    Summary: gets the subset of a dataframe by the query of interest
    Args:
        query (str): search term of interest, use [space] for OR & | [pipe] for AND
        dfCounts (pandas.DataFrame): counts dataframe
    Returns:
        geneSorted (pandas.DataFrame): dataframe for genes containing query
    """

    queryList = query.replace(" ", "|").replace("+", " ")
    geneSorted = dfCounts[dfCounts[level] == queryList]
    return geneSorted

def subset_by_taxa(studyName, level, query, testMerge):
    """
    Summary: subsets counts dataframe by query (gene name/gene family name)
    Args:
        studyName (str): name of study directory
        level (str): taxonomic level of interest
        testMerge (pandas.DataFrame): merged annotated counts
    Returns:
        geneDF (pandas.DataFrame): dataframe of counts only for gene/query of interest
    """

    os.chdir(studyName)
    geneDF = search_by_taxa(query, level, testMerge)
    geneDF.to_csv(str(query)+'_'+'allcounts.csv')
    os.chdir('..')
    return geneDF





import glob
import math
import numpy as np
import pandas as pd
import sys
import os
import errno
from skbio.stats import subsample_counts
from functools import reduce
from Bio import Entrez
import time
import re
from scipy import stats
import pkg_resources
pd.options.mode.chained_assignment = None
############################################################################################

def convert_annotation_to_counts(tableName, groupName, sampleID):
    """
    Summary: converts annotation files gotten from API to non-normalised counts data for both function and organism

    Args:
        tableName (str): sample's annotation table as gotten from MG-RAST API
        groupName (str): sample's group
        sampleID  (str): sample's ID


    Returns:
        counts (pandas.DataFrame): pandas DataFrame with columns  'gene function', 'organism', and 'count'
    """
    print('reading ' + str(tableName))
    rawData = pd.read_table(tableName)
    rawData['organism'] = rawData['semicolon separated list of annotations'].str.extract("org\w+\=\[([^\]]+)\]", expand = False)
    rawData['gene function'] = rawData['semicolon separated list of annotations'].str.extract("func\w+\=\[([^\]]+)\]", expand = False)
    print('filtering by max percentage identity...')
    rawData = keep_max_pctID_only(rawData)
    counts = get_and_save_counts(rawData, groupName, sampleID)
    return counts

def get_group_counts_API(studyName):
    """
    Summary: converts all annotations in a study, with group awareness, to raw counts data


    Args:
        studyName (str): name of the directory under which annotations are stored

    Returns:
        grpsCountsList (list): list containing each counts dataFrame for each sample. This is also output to CSV format in a folder 'counts'

    """
    os.chdir(str(studyName))
    print('loading files in ' + studyName)
    metagenomeList = glob.glob('*.tsv')
    filenames = [metagenome.split('.')[0] for metagenome in metagenomeList]
    groupName = [name.split('_')[0] for name in filenames]
    sampleID = [name.split('_')[1] for name in filenames]
    print('converting annotations to counts...')
    grpCountsList = [convert_annotation_to_counts(metagenomeList[i], groupName[i], sampleID[i]) for i in range(len(metagenomeList))]
    os.chdir('..')
    return grpCountsList

############################################################################################
'''
TRANSFORMING ANNOTATIONS INTO NORMALISED COUNTS DATA











'''
############################################################################################

####CREATE PATH IN DIRECTORY##
def create_path(path):
    """
    Summary: creates a path, with error checking

    Args:
        path (str): name of folder/directory to be created

    Returns:
        None, creates folder
    """

    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise

####READ CSV FILE INTO PANDAS DATAFRAME####
def make_df(filename):
    """
    Summary:

    Args:

    Returns:
    """

    #df = pd.read_csv(filename, header=None)
    #return df

    df = pd.read_csv(filename) #index_col=[0,1]
    return df


###KEEP ANNOTATIONS WITH MAX PERCENT IDENTITY###
def keep_max_pctID_only(table):
    """
    Summary: filters annotations by maximum percentage identity

    Args:
        table (pandas.DataFrame): annotations table, as DataFrame

    Returns:
        maxPctID (pandas.DataFrame): filtered annotations table, as DataFrame
    """
    mask = table.groupby('query sequence id').agg('idxmax')
    maxPctID = table.loc[mask['percentage identity']].reset_index()
    return maxPctID

###MERGE FUNCTION AND ORGANISM ANNOTATION TABLES###
def merge_annotation_tables(table1, table2):
    """
    Summary:

    Args:

    Returns:
    """
    mergedTable = pd.merge(table1, table2, on = 'query sequence id').sort_values(['semicolon separated list of annotations_x', 'semicolon separated list of annotations_y'], ascending = [0,0])
    mergedTable.rename(columns={'semicolon separated list of annotations_x':'gene function','semicolon separated list of annotations_y':'organism'})
    groupByQuery = mergedTable.groupby('query sequence id')
    return mergedTable

###GET RAW COUNTS FOR MERGED ANNOTATION TABLES###
def get_and_save_counts(mergedTable, groupName, sampleID):
    """
    Summary:

    Args:

    Returns:
    """
    print('sorting values...')
    counts1 = mergedTable.sort_values(['gene function', 'organism']).groupby(['gene function', 'organism']).size()
    newFileName = '%(group)s_%(sample)s_counts.csv' % \
            {"group": groupName, "sample": sampleID}
    print('writing ' + newFileName  +  ' to file')
    counts1.to_csv(newFileName)
    print('written ' + str(newFileName))
    #counts = pd.read_csv(newFileName)
    counts1.columns = ['gene function', 'organism', groupName + '_' + sampleID]
    return counts1

####GO FROM TWO TABLES (FXN AND ORG) TO COUNTS####
def counts_within_metagenome(fxnTable, orgTable, groupName, sampleID):
    """
    Summary:

    Args:

    Returns:
    """
    mergedTable = merge_annotation_tables(fxnTable, orgTable)
    counts = get_and_save_counts(mergedTable, groupName, sampleID)
    return counts


#####CONVERTING TAB ANNOTATIONS FROM MG-RAST INTO RAW COUNTS DATA#####
def get_counts_within_metagenome(fxnTable, orgTable, groupName, sampleID):
    """
    Summary:

    Args:

    Returns:
    """
    fxnData = pd.read_csv(fxnTable, delimiter = '\t', usecols=['query sequence id', 'percentage identity', 'bit score', 'semicolon separated list of annotations'])
    orgData = pd.read_csv(orgTable, delimiter = '\t', usecols=['query sequence id', 'percentage identity', 'bit score', 'semicolon separated list of annotations'])

    fxnDataMax = keep_max_pctID_only(fxnData)
    orgDataMax = keep_max_pctID_only(orgData)

    counts = counts_within_metagenome(fxnDataMax, orgDataMax, groupName, sampleID)

    return counts

###CONVERT TAB ANNOTATIONS TO COUNTS WITHIN A GROUP####
def get_counts_within_group(studyName, groupName):
    """
    Summary:

    Args:

    Returns:
    """
    fxnDataList = glob.glob(os.path.join(studyName + '/' + groupName, '*function*.tab'))
    orgDataList = glob.glob(os.path.join(studyName + '/' + groupName, '*organism*.tab'))
    grpCountsList = [get_counts_within_metagenome(fxnDataList[i], orgDataList[i], groupName, str(i+1)) for i in range(len(fxnDataList))]
    return grpCountsList

#####GET ALL THE COUNTS (FROM TAB ANNOTATIONS) OF ALL THE GROUPS IN A STUDY#####
def get_all_group_counts(studyName):
    """
    Summary:

    Args:

    Returns:
    """
    grpsList = [poo.split('/')[1] for poo in glob.glob(os.path.join(studyName, '*'))]
    fullCountsList = [get_counts_within_group(studyName, grp) for grp in grpsList]
    return fullCountsList
########################################NORMALISATION###############################################
'''Normalisation by subsampling and recodification'''
###FIND SAMPLING DEPTH FOR SUBSAMPLING###
def find_sampling_depth(rawCounts):
    """
    Summary: determines median counts of all samples and sets as sampling depth

    Args:
        rawCounts (list): list of all raw counts DataFrames, each corresponding to a sample

    Returns:
        samplingDepth (int): median number of counts across all samples
    """
    sums = []
    print('finding sampling depth...')
    for i in range(len(rawCounts)):
        rawCounts[i] = rawCounts[i].set_index(['gene function', 'organism'])
        sums.insert(i, int(rawCounts[i].sum()))
    lenSeries = pd.Series(sums)
    samplingDepth = int(lenSeries.median())
    return samplingDepth

####WRITE RAREFIED AND RECODED COUNTS TO FILE####
def rarefy_and_recode(filenames, rawCounts, samplingDepth):
    """
    Summary: subsamples all samples to the median (average of 100 times),  

    Args:
        filenames ()
        rawCounts ()
        samplingDepth ()

    Returns:
    """
    for i in range(len(rawCounts)):
        subsampleList = []
        if int(rawCounts[i].sum()) < samplingDepth:
            meanSubsample = rawCounts[i]
        else:
            for j in range(100):
                sample = subsample_counts(rawCounts[i].transpose().values[0], samplingDepth)
                subsampleList.insert(j, sample)
            print("completed 100 subsamples for sample number " + str(i))
            meanSubsample = pd.Series(subsampleList).mean()
            #recodification: setting all values less than 1.01 to zero
            meanSubsample[meanSubsample < 1.01] = 0
        sampleName = filenames[i].split('.')[0]
        rawCounts[i][sampleName] = meanSubsample
        newFileName = sampleName + "_norm.csv"
        create_path('normalised_counts')
        rawCounts[i].to_csv(os.path.join('normalised_counts', newFileName))
        print("written " + newFileName + " to file.")
    return

#####NORMALISE BY R&R (TO MEDIAN)#####
def normalise_rnr(studyName):
    """
    Summary: normalises all counts in study by rarefaction and recodification

    Args:
        studyName (str): name of directory/study containing raw non-normalised counts

    Returns: None, outputs normalised counts into studyName/normalised_counts
    """
    os.chdir(str(studyName))
    print('loading files in ' + studyName)
    metagenomeList = glob.glob('*counts.csv')
    flatCountsList = [pd.read_csv(metagenome, names=['gene function', 'organism', 'counts']) for metagenome in metagenomeList]
    #flatCountsList = sum(fullCountsList, [])
    samplingDepth = find_sampling_depth(flatCountsList)
    filenames = glob.glob(os.path.join('counts/*.csv'))
    rarefy_and_recode(metagenomeList, flatCountsList, samplingDepth)
    os.chdir('..')
    return
############################################################################################
'''
ANNOTATING COUNTS DATA

Communicates with Entrez E-utilities to annotate hierarchies









'''
############################################################################################
''' Insert the proper version of the get_taxid function '''
def get_taxid(species):
    """
    Summary: gets Entrex taxid for given lineage name 

    Args:
        species (str): species, genus, etc

    Returns:
        (str): taxid for given name
    """
    species = species.split(';')[0].replace(" ", "+").replace("''", " ").replace(":", " ").replace("(", " ").replace(")", " ")#.split('+')
    search = Entrez.esearch(term = species, db = "taxonomy", retmode = "xml", usehistory = "n")
    record = Entrez.read(search)
    if (record['IdList'] == []):
        name_fixed = record['ErrorList']['PhraseNotFound'][0]#.split("+")
        #redo = get_taxid(name_fixed[0] + " " + name_fixed[1])  ''' make this function non-recursive'''
        search2 = Entrez.esearch(term = name_fixed, db = "taxonomy", retmode = "xml")
        record = Entrez.read(search2)
        if (record['IdList'] == []):
            thirdTry = species.split("+")[0] + " " + species.split("+")[1]
            search3 = Entrez.esearch(term = thirdTry, db = "taxonomy", retmode = "xml")
            record = Entrez.read(search3)
        else:
            record = record
    else:
        record = record

    return record['IdList'][0]

def get_tax_data(taxid):
    """
    Summary: access lineage in Entrez database for given taxid

    Args:
        taxid (str): the taxid for which lineage information is desired

    Returns:
        JSON entry for Entrez taxid

    """
    search = Entrez.efetch(id = taxid, db = "taxonomy", retmode = "xml")
    return Entrez.read(search)


def get_tax_hierarchy(species):
    """
    Summary: return taxonomic hierarchy of species

    Args:
        species (str): species, genus, etc for which lineage is desired

    Returns:
        lineage (dict): dictionary with {taxonomic level : name}
    """
    try:
        taxid = get_taxid(species)
        data = get_tax_data(taxid)
    except:
        time.sleep(50)
        taxid = get_taxid(species)
        data = get_tax_data(taxid)
    lineage = {d['Rank']:d['ScientificName'] for d in data[0]['LineageEx'] if d['Rank'] in ['phylum', 'class', 'order', 'family', 'genus']}
    return lineage





#HIERARCHY ANNOTATION#
#input is a series of species names







def get_taxonomic_hierarchy(organismSeries, userEmail):
    """
    Summary: gets all the taxonomic hierarchies/lineages for a series of organism names

    Args: 
        organismSeries (pandas.Series): Series containing  organisms
        userEmail (string): email of tool developer

    Returns:
    """

    Entrez.email = userEmail
    Entrez.tool = "Vibhu-species2taxon"
    if not Entrez.email:
        print("you must add your email address")
        sys.exit(2)

    lineage_list = []

    print('parsing taxonomic data...') #starting the parser

    for species in organismSeries:
        print ('\t' + species.split(";")[0]) #print progress
        lineage = get_tax_hierarchy(species)
        lineage_list.append(str(lineage))
        time.sleep(5)

    return lineage_list


##ANNOTATE ENTIRE METAGENOMES DATA FRAME WITH HIERARCHY##
#input is a dataframe in which the numerical columns are each samples
def annotate_with_hierarchy(dfCounts):
    """
    Summary: annotates normalised counts with lineage for each organism

    Args:
        dfCounts (pandas.DataFrame): counts dataframe

    Returns:
        dfCounts (pandas.DataFrame): counts dataframe with columns for taxonomic hierarchy
    """
    io = pkg_resources.resource_stream("metadex", "lineage_dict.csv")
    dfCounts['lineage'] = dfCounts['organism'].str.split(";").str[0]

    speciesList = dfCounts['lineage'].unique().tolist()
    speciesSeries = pd.Series(speciesList)
    lineageDF = pd.read_csv(io)
    dictSpeciesList = lineageDF['species'].tolist()
    diffList = list(set(speciesList)-set(dictSpeciesList))
    initlineageDict = dict(zip(lineageDF['species'].tolist(), lineageDF['lineage'].tolist()))



    lineageList = get_taxonomic_hierarchy(pd.Series(diffList), 'vibhuc@me.com')

    difflineageDict = dict(zip(diffList, lineageList))
    lineageDict = initlineageDict.copy()
    lineageDict.update(difflineageDict)
    lineageDF = pd.DataFrame({'name':lineageDict.keys(), 'lineage':lineageDict.values()})
    lineageDF.to_csv('updated_lineage_dict.csv')
    dfCounts['lineage'].replace(lineageDict, inplace = True)

    dfCounts['phylum'] = dfCounts['lineage'].str.extract("'p\w+\'\: \'([A-z]\w+)\'", expand = False)
    dfCounts['class'] = dfCounts['lineage'].str.extract("'c\w+\'\: \'([A-z]\w+)\'", expand = False)
    dfCounts['order'] = dfCounts['lineage'].str.extract("'o\w+\'\: \'([A-z]\w+)\'", expand = False)
    dfCounts['family'] = dfCounts['lineage'].str.extract("'f\w+\'\: \'([A-z]\w+)\'", expand = False)
    dfCounts.pop('lineage')
    return dfCounts

def annotate_all_counts(studyName):
    """
    Summary: annotates all counts within a study with taxonomic hierarchy

    Args:
        studyName (str): name of directory within which counts are located

    Returns:
        None, outputs annotated files to norm_taxonomy folder
    """

    normFilesList = glob.glob(os.path.join(studyName + '/normalised_counts', '*norm.csv'))
    normFileNames = [normFile.strip().split('/')[-1].split('.')[0] for normFile in normFilesList]
    normDFsList = [pd.read_csv(normCount) for normCount in normFilesList]
    [normDF.pop('counts') for normDF in normDFsList]
    annotatedList = [annotate_with_hierarchy(normDF) for normDF in normDFsList]
    os.chdir(studyName)
    os.mkdir('norm_taxonomy')
    os.chdir('norm_taxonomy')
    [normDFsList[i].to_csv(normFileNames[i] + '_taxonomy.csv') for i in range(len(normDFsList))]
    return annotatedList



##############################PRE-PROCESSING FOR GENE-ONLY COUNTS MATRIX##############################
###RUN BORUTA TO DETERMINE WHICH GENES HAVE CHANGES MOST###
import pandas as pd
from sklearn.ensemble import RandomForestClassifier
from boruta import BorutaPy
'''
# load X and y
# NOTE BorutaPy accepts numpy arrays only, hence the .values attribute
X = pd.read_csv('examples/test_X.csv', index_col=0).values
y = pd.read_csv('examples/test_y.csv', header=None, index_col=0).values
y = y.ravel()

# define random forest classifier, with utilising all cores and
# sampling in proportion to y labels
rf = RandomForestClassifier(n_jobs=-1, class_weight='auto', max_depth=5)

# define Boruta feature selection method
feat_selector = BorutaPy(rf, n_estimators='auto', verbose=2, random_state=1)

# find all relevant features - 5 features should be selected
feat_selector.fit(X, y)

# check selected features - first 5 features are selected
feat_selector.support_

# check ranking of features
feat_selector.ranking_

# call transform() on X to filter it down to selected features
X_filtered = feat_selector.transform(X)

#write the filtered genes counts matrix to file for graphical analysis#
'''
def find_genes_of_interest(studyName, groupsList, geneCounts, lvl1pct=70, lvl2pct=70, lvl3pct=60, fxnpct=40):
    """
    Summary: uses Boruta machine learning method to roughly determine potential genes of interest. requires tab-separated  matrix from MG-RAST analysis page

    Args:
        studyName (str): directory (study name)
        geneCountsName (str): filename for tab separated matrix
        lvl1pct (int): threshold for Boruta on level 1
        lvl2pct (int): threshold for Boruta on level 2
        lvl3pct (int): threshold for Boruta on level 3
        fxnpct (int): threshold for Boruta on gene name


    Returns: None, outputs files with tentative genes/gene families of interest

    """

    #geneCounts = pd.read_table(geneCountsName, header=0)#, header=0)#header=0
    numGeneCounts = geneCounts.select_dtypes(include=[np.number])
    samplingDepth = numGeneCounts.sum().median()
    os.chdir(studyName)
    for i in range(len(numGeneCounts.columns)):
        subsampleList = []
        if int(numGeneCounts[numGeneCounts.columns[i]].sum()) < samplingDepth:
            meanSubsample = numGeneCounts[numGeneCounts.columns[i]]
        else:
            for j in range(100):
                sample = subsample_counts(numGeneCounts[numGeneCounts.columns[i]].transpose().values, int(samplingDepth))
                subsampleList.insert(j, sample)
            print("completed 100 subsamples for sample number " + str(i))
            meanSubsample = pd.Series(subsampleList).mean()
            #recodification: setting all values less than 1.01 to zero
            meanSubsample[meanSubsample < 1.01] = 0
        meanSubsample = meanSubsample/meanSubsample.sum()
        numGeneCounts[numGeneCounts.columns[i]] = meanSubsample
    countsLvl1 = geneCounts.groupby('level1').sum()
    countsLvl2 = geneCounts.groupby('level2').sum()
    countsLvl3 = geneCounts.groupby('level3').sum()
    countsLvl4 = geneCounts.groupby('function').sum()
    levelList = [countsLvl1, countsLvl2, countsLvl3, countsLvl4]
    countsLvl1.to_csv(studyName + 'genes_lvl1.csv')
    countsLvl2.to_csv(studyName + 'genes_lvl2.csv')
    countsLvl3.to_csv(studyName + 'genes_lvl3.csv')
    countsLvl4.to_csv(studyName + 'genes_function.csv')
    groupsDict = dict(enumerate(pd.Series(groupsList).unique()))
    dictGroups = {y:x for x,y in groupsDict.items()}
    Y = pd.Series(groupsList).replace(dictGroups).values
    rf = RandomForestClassifier(n_jobs = -1, class_weight = 'auto', max_depth=3)

    X = countsLvl1.transpose().values
    feat_selector = BorutaPy(rf, n_estimators='auto', verbose=2, perc=int(lvl1pct))
    feat_selector.fit(X, Y)
    if len(countsLvl1[feat_selector.support_]) > 0:
        countsLvl1[feat_selector.support_].to_csv('level1_tentative.csv')
    countsLvl1[feat_selector.support_weak_].to_csv('level1_tentative_weak.csv')

    X = countsLvl2.transpose().values
    feat_selector = BorutaPy(rf, n_estimators='auto', verbose=2, perc=int(lvl2pct))
    feat_selector.fit(X, Y)
    if len(countsLvl2[feat_selector.support_]) > 0:
        countsLvl2[feat_selector.support_].to_csv('level2_tentative.csv')
    countsLvl2[feat_selector.support_weak_].to_csv('level2_tentative_weak.csv')

    X = countsLvl3.transpose().values
    feat_selector = BorutaPy(rf, n_estimators='auto', verbose=2, perc=int(lvl3pct))
    feat_selector.fit(X, Y)
    if len(countsLvl3[feat_selector.support_]) > 0:
        countsLvl3[feat_selector.support_].to_csv('level3_tentative.csv')
    countsLvl3[feat_selector.support_weak_].to_csv('level3_tentative_weak.csv')

    X = countsLvl4.transpose().values
    feat_selector = BorutaPy(rf, n_estimators='auto', verbose=2, perc=int(fxnpct))
    feat_selector.fit(X, Y)
    if len(countsLvl4[feat_selector.support_]) > 0:
        countsLvl4[feat_selector.support_].to_csv('function_tentative_.csv')
    countsLvl4[feat_selector.support_weak_].to_csv('function_tentative_weak.csv')
    os.chdir('..')

def merge_dfs1(ldf, rdf):
    """
    Summary: merges two counts (with hierarchy)
    Args:
        ldf (pandas.DataFrame): left dataframe
        rdf (pandas.DataFrame): right dataframe
    Returns:
        pd.merge(ldf, rdf) (pandas.DataFrame): merged dataframe
    """

    print('yep, again...')
    return ldf.merge(rdf, how='outer', on=['gene function', 'organism', 'phylum', 'class', 'order', 'family'])

def merge_all_counts2(studyName):
    """
    Summary: merges all counts w/ hierarchy
    Args:
        studyName (str): study name/directory name (ideally one and the same)
    Returns:
        mergedCounts (pandas.DataFrame): merged dataframe
    """

    annotatedCountsList = glob.glob(studyName + '/norm_taxonomy/*norm_taxonomy.csv')
    annotatedDFsList = [pd.read_csv(count) for count in annotatedCountsList]
    [annDF.pop('Unnamed: 0') for annDF in annotatedDFsList]
    mergedCounts  = reduce(merge_dfs1, annotatedDFsList)
    mergedCounts.fillna(0)
    mergedCounts.to_csv(str(studyName) + '_allcounts.csv')
    return mergedCounts


def find_organisms_of_interest(studyName, groupsList,  lvl1pct=70, lvl2pct=70, lvl3pct=60, lvl4pct=40):
    """
    Summary: uses Boruta machine learning method to roughly determine potential genes of interest. requires tab-separated  matrix from MG-RAST analysis page

    Args:
        studyName (str): directory (study name)
        geneCountsName (str): filename for tab separated matrix
        lvl1pct (int): threshold for Boruta on level 1
        lvl2pct (int): threshold for Boruta on level 2
        lvl3pct (int): threshold for Boruta on level 3
        fxnpct (int): threshold for Boruta on gene name


    Returns: None, outputs files with tentative genes/gene families of interest

    """

    #geneCounts = pd.read_table(geneCountsName, header=0)#, header=0)#header=0
    mergedCounts = merge_all_counts2(studyName)
    mergedCounts = mergedCounts.fillna(0)
    countsLvl1 = mergedCounts.groupby('phylum').sum()
    countsLvl2 = mergedCounts.groupby('class').sum()
    countsLvl3 = mergedCounts.groupby('order').sum()
    countsLvl4 = mergedCounts.groupby('family').sum()
    levelList = [countsLvl1, countsLvl2, countsLvl3, countsLvl4]
    countsLvl1.to_csv(studyName + '_phylum.csv')
    countsLvl2.to_csv(studyName + '_class.csv')
    countsLvl3.to_csv(studyName + '_order.csv')
    countsLvl4.to_csv(studyName + '_family.csv')
    groupsDict = dict(enumerate(pd.Series(groupsList).unique()))
    dictGroups = {y:x for x,y in groupsDict.items()}
    Y = pd.Series(groupsList).replace(dictGroups).values
    rf = RandomForestClassifier(n_jobs = -1, class_weight = 'auto', max_depth=3)

    X = countsLvl1.transpose().values
    feat_selector = BorutaPy(rf, n_estimators='auto', verbose=2, perc=int(lvl1pct))
    feat_selector.fit(X, Y)
    if len(countsLvl1[feat_selector.support_]) > 0:
        countsLvl1[feat_selector.support_].to_csv('phylum_tentative.csv')
    countsLvl1[feat_selector.support_weak_].to_csv('phylum_tentative_weak.csv')

    X = countsLvl2.transpose().values
    feat_selector = BorutaPy(rf, n_estimators='auto', verbose=2, perc=int(lvl2pct))
    feat_selector.fit(X, Y)
    if len(countsLvl2[feat_selector.support_]) > 0:
        countsLvl2[feat_selector.support_].to_csv('class_tentative.csv')
    countsLvl2[feat_selector.support_weak_].to_csv('class_tentative_weak.csv')

    X = countsLvl3.transpose().values
    feat_selector = BorutaPy(rf, n_estimators='auto', verbose=2, perc=int(lvl3pct))
    feat_selector.fit(X, Y)
    if len(countsLvl3[feat_selector.support_]) > 0:
        countsLvl3[feat_selector.support_].to_csv('order_tentative.csv')
    countsLvl3[feat_selector.support_weak_].to_csv('order_tentative_weak.csv')

    X = countsLvl4.transpose().values
    feat_selector = BorutaPy(rf, n_estimators='auto', verbose=2, perc=int(fxnpct))
    feat_selector.fit(X, Y)
    if len(countsLvl4[feat_selector.support_]) > 0:
        countsLvl4[feat_selector.support_].to_csv('family_tentative_.csv')
    countsLvl4[feat_selector.support_weak_].to_csv('family_tentative_weak.csv')
    os.chdir('..')





from .preprocessing import *
from .analysis import *
from pkg_resources import resource_string

class Study:

    def __init__(self, name=None, counts=None, groups=None, iDs=None, geneMatrix=None):
        self.name = name

    def read_annotations(self):
        self.counts = get_group_counts_API(self.name)
        filenames = glob.glob(self.name +'/*.csv')
        self.groups = [filename.split('/')[1].split('_')[0] for filename in filenames]
        self.iDs = [filename.split('/')[1].split('_')[1] for filename in filenames]



    def normalise_counts(self):
        normalise_rnr(self.name)
        normalisedList = glob.glob(os.path.join('normalised_counts/*.csv'))
        normalisedDFList = [pd.read_csv(norm) for norm in normalisedList]
        self.counts = normalisedDFList

    def load_counts(self, folder):
        filenames = glob.glob(folder +'/*.csv')
        self.counts = [pd.read_csv(filename) for filename in filenames]
        self.groups = [filename.split('/')[1].split('_')[0] for filename in filenames]
        self.iDs = [filename.split('/')[1].split('_')[1] for filename in filenames]

    def add_lineage_info(self):
        self.counts = annotate_all_counts(self.name)


    def load_gene_matrix(self, filename):
        self.geneMatrix = pd.read_table(filename, header=0)

    def determine_genes_of_interest(self, lvl1pct=70, lvl2pct=70, lvl3pct=60, fxnpct=40):
        find_genes_of_interest(self.name, self.groups, self.geneMatrix, lvl1pct, lvl2pct, lvl3pct, fxnpct)
        return [pd.read_csv(i) for i in glob.glob('*tentative*.csv')]

    def determine_organisms_of_interest(self, lvl1pct=70, lvl2pct=70, lvl3pct=60, lvl4pct=40):
        find_organisms_of_interest(self.name, self.groups, lvl1pct, lvl2pct, lvl3pct, lvl4pct)
        return [pd.read_csv(i) for i in glob.glob('*tentative*.csv')]


    def focus_on_gene(self, searchString):
        mergedCounts = merge_all_counts(self.name)
        mergedCounts = mergedCounts.fillna(0)
        geneDF = subset_by_gene(self.name, searchString, mergedCounts)
        return geneDF

    def visualise_diversity_for_gene(self, geneName, searchString):
        get_diversity_measures(geneName, searchString, self.name)

    def focus_on_organisms(self, level, query):
        mergedCounts = merge_all_counts(self.name)
        mergedCounts = mergedCounts.fillna(0)
        orgDF = subset_by_taxa(self.name, level, query, mergedCounts)
        return orgDF
